package ServerConfig;

import BankApp.*;
import BankServer.*;
import ServerConfig.loader.*;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.net.InetAddress;

import java.util.Properties;
import org.omg.CORBA.*;
import org.omg.CosNaming.*;
import org.omg.CosNaming.NamingContextPackage.*;

/**
 * @author Min Huang (khtwo)
 * ServerConfig is used to store server or monitor configurations
 * It provides functions to operation on central naming service, to build naming structures
 * and to retrieve related information when needed
 */
public class ServerConfig {
	// Directory structure
	public static String PrimaryTag = "Primary";
	public static String branchDirectory = "Branches";
	public static String branchTag = "Branch";
	public static String serverTag = "Server";
	public static String monitorDirectory = "Monitors";
	public static String monitorTag = "Monitor";
	public static String loaderTag = "loader";
	public static String uniqueTag = "Unique";
	public static final int semaphoreRetry = 3;

	public NamingContextExt rootContext;	// The root naming context
	
	// Config file
	public String centralNSConfigFileName = "centralNS.cfg";
	public String serverConfigFileName = "server.cfg";
	public String monitorConfigFileName = "monitor.cfg";
	public boolean isServer = true;		// If this config for a server
	public boolean isMonitor = false;	// If this config for a monitor

	// Naming Context server configure
	public static String centralNCInitialHost = ""; // Host address of central Naming Context server
	public static String centralNCInitialPort = ""; // Initial service port of central Naming Context server
	public static String centralNCPort = ""; // Service port of central Naming Context server
	public String localNCInitialPort = ""; // Initial service port of local Naming Context server
	public String localNCPort = ""; // Service port of local Naming Context server

	// Common Configure
	public String localhostIP = ""; // IP address used to provide service on localhost

	// Primary and backup server configure
	public String branchID = ""; // Branch ID that managed by this server
	public String serverID = ""; // Branch ID that managed by this server
	public String dataFileName = ""; // The name of the data file = "data" + branchID + ".dat"
	public String logFileName = ""; // The name of the log file = "log" + branchID + ".log"
	public String UDPPort = ""; // The UDP port for primary server UDP listener
	public String backupUDPPort = ""; // The UDP port for backup server UDP listener
	public String transferPort = ""; // The TCP port for transfer operation

	// Monitor configure
	public String monitorID = ""; // Branch ID that managed by this server

	// Program Loader
	public String currentDirectory = "";	// The working directory
	public ProgramLoader loader;	// The reference of the loader object
	public String serverLoadingCommand = "";	// The loading command for the program
	public String monitorLoadingCommand = "";	// The loading command for the program
	
	// Orb
	public ORB orb = null;
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// Check running arguments, give out usage description if the arguments are not correct
		if(args.length<2){
			System.out.println("Usage:");
			System.out.println("ServerConfig list/erase/load all/<branchID>/<monitorID>");
			return;
		}
		if(args[0].compareToIgnoreCase("list")!=0 && args[0].compareToIgnoreCase("erase")!=0 && args[0].compareToIgnoreCase("load")!=0){
			System.out.println("Usage:");
			System.out.println("ServerConfig list/erase/load all/<branchID>/<monitorID>");
			return;
		}
		
		ServerConfig currentConfig = new ServerConfig();
		try {
			// Try to read server config
			try{
				currentConfig.readNSFromFile();
			}catch(Exception e){
				// No config is read, do nothing and quit
				System.out.println("Can not read central naming service configure file.");
				Runtime.getRuntime().exit(1);					
			}

			// Use the central naming service
			Properties props = new Properties();
			props.put("org.omg.CORBA.ORBInitialPort", ServerConfig.centralNCInitialPort);
			props.put("org.omg.CORBA.ORBInitialHost", ServerConfig.centralNCInitialHost);
			String empty[] = new String[0];
			ORB orb = ORB.init(empty, props);
			currentConfig.setServerConfig(orb);

			// Retrieve root context
			NamingContextExt rootContext = currentConfig.rootContext;

			// Perform the functions accordingly
			
			if(args[0].compareToIgnoreCase("list") == 0){
				// Command "list"
				System.out.println("Listing server and monitor naming context information");
				if(args[1].compareToIgnoreCase("all") == 0){
					// Print all information in the root naming context
					System.out.println("<NameService>");
					ServerConfig.printNamingDirectory(rootContext, "  ", false);
				} else if(args[1].length()==3){
					// Print branch information
					try{
						NamingContextExt ctx2 = NamingContextExtHelper.narrow(rootContext.resolve_str(ServerConfig.branchDirectory));
						ctx2 = NamingContextExtHelper.narrow(ctx2.resolve_str(ServerConfig.branchTag + args[1]));
						System.out.println("<NameService><Branches><Branch"+args[1]+">");
						ServerConfig.printNamingDirectory(ctx2, "  ", false);
					}catch(Exception e){
						// Finding branch failed, quit
						System.out.println("Branch not found.");
					}
				} else{
					// Print monitor information
					try{
					// Print monitor information
						NamingContextExt ctx2 = NamingContextExtHelper.narrow(rootContext.resolve_str(ServerConfig.monitorDirectory));
						ctx2 = NamingContextExtHelper.narrow(ctx2.resolve_str(ServerConfig.monitorTag + args[1]));
						System.out.println("<NameService><Monitors><Monitor"+args[1]+">");
						ServerConfig.printNamingDirectory(ctx2, "  ", false);
					}catch(Exception e){
						// Finding branch failed, quit
						System.out.println("Monitor not found.");
					}
				}
				return;
			}
			
			if(args[0].compareToIgnoreCase("erase") == 0){
				// Command "erase"
				System.out.println("Deleting server and monitor naming context information");
				if(args[1].compareToIgnoreCase("all") == 0){
					// Print all information in the root naming context and erase them
					System.out.println("<NameService>");
					ServerConfig.printNamingDirectory(rootContext, "  ", true);
				} else if(args[1].length()==3){
					// Erase branch information
					try{
						// Get into that directory
						NamingContextExt ctx2 = NamingContextExtHelper.narrow(rootContext.resolve_str(ServerConfig.branchDirectory));
						NamingContextExt ctx3 = NamingContextExtHelper.narrow(ctx2.resolve_str(ServerConfig.branchTag + args[1]));
						System.out.println("<NameService><Branches><Branch"+args[1]+">");
						// Erase the contents in the branch directory
						ServerConfig.printNamingDirectory(ctx3, "  ", true);
						// Erase the branch directory itself
						NameComponent name[] = ctx2.to_name(ServerConfig.branchTag + args[1]);
						ctx2.unbind(name);
					}catch(Exception e){
						// Finding branch failed, quit
						System.out.println("Branch not found.");
					}
				} else{
					// Erase monitor information
					try{
						// Get into that directory
						NamingContextExt ctx2 = NamingContextExtHelper.narrow(rootContext.resolve_str(ServerConfig.monitorDirectory));
						NamingContextExt ctx3 = NamingContextExtHelper.narrow(ctx2.resolve_str(ServerConfig.monitorTag + args[1]));
						System.out.println("<NameService><Branches><Branch"+args[1]+">");
						// Erase the contents in the monitor directory
						ServerConfig.printNamingDirectory(ctx3, "  ", true);
						// Erase the monitor directory itself
						NameComponent name[] = ctx2.to_name(ServerConfig.branchTag + args[1]);
						ctx2.unbind(name);
					}catch(Exception e){
						// Finding branch failed, quit
						System.out.println("Monitor not found.");
					}
				}
				return;
			}
			
			if(args[0].compareToIgnoreCase("load") == 0){
				// Command "load"
				System.out.println("Loading server and monitor program");
				if(args[1].compareToIgnoreCase("all") == 0){
					// Load all servers and monitors that registered in this system

					// Find current registered servers
					String serverList[] = ServerConfig.listServers(rootContext);
					// Load those servers one by one
					for(int i=0; i<serverList.length; i++){
						System.out.print("Loading Server " + serverList[i] + ".  ");
						// Retrieve current branchID and serverID
						String branchID = serverList[i].substring(6,9);
						String serverID = serverList[i].substring(16,serverList[i].length());

						// Try to launch the server
						try{
							// Get the server configure and load the server
							ServerConfig server = ServerConfig.getServerConfig(rootContext, branchID, serverID);
							server.relaunch();
							System.out.println("ok");
						}
						catch(Exception e){
							System.out.println("Failed");
						}
					}

					// Find current monitors
					String monitorList[] = ServerConfig.listMonitors(rootContext);
					// Load those monitors one by one
					for(int i=0; i<monitorList.length; i++){
						// Retrieve current monitorID
						System.out.print("Loading monitor " + monitorList[i] + ".  ");
						String monitorID = monitorList[i].substring(7,monitorList[i].length());

						// Try to launch the monitor
						try{
							// Get the monitor configure and load the monitor
							ServerConfig monitor = ServerConfig.getMonitorConfig(rootContext, monitorID);
							monitor.relaunch();
							System.out.println("ok");
						}
						catch(Exception e){
							System.out.println("Failed");
						}
					}
				} // Enf of loading all servers and monitors
				else if(args[1].length()==3){
					// Load servers in a specific branch
					
					// Load server1
					System.out.print("Loading Server " + args[1] + "\\Server1.  ");
					// Try to launch the server
					try{
						// Get the server configure and load the server
						ServerConfig server = ServerConfig.getServerConfig(rootContext, args[1], "1");
						server.relaunch();
						System.out.println("ok");
					}
					catch(Exception e){
						System.out.println("Failed");
					}
					// Load server2
					System.out.print("Loading Server " + args[1] + "\\Server2.  ");
					// Try to launch the server
					try{
						// Get the server configure and load the server
						ServerConfig server = ServerConfig.getServerConfig(rootContext, args[1], "2");
						server.relaunch();
						System.out.println("ok");
					}
					catch(Exception e){
						System.out.println("Failed");
					}
				} // End of loading servers in a branch
				else{
					// Load a specific monitor
					System.out.print("Loading Monitor" + args[1] + ".  ");
					// Try to launch the monitor
					try{
						// Get the server configure and load the server
						ServerConfig monitor = ServerConfig.getMonitorConfig(rootContext, args[1]);
						monitor.relaunch();
						System.out.println("ok");
					}
					catch(Exception e){
						System.out.println("Failed");
					}
				} // End of loading monitor
			}

			
/*			
			if(args.length == 3 || args.length == 1){
				String command = "";
				if(args.length == 3){
					command = args[2];
				}else{
					command = args[0];
				}
				if(command.compareToIgnoreCase("erase")==0){
					toErase = true;
					System.out.println("<NameService>");
					ServerConfig.printNamingDirectory(rootContext, "  ", toErase);
				} else if(command.compareToIgnoreCase("list")==0){
					toErase = false;
					System.out.println("<NameService>");
					ServerConfig.printNamingDirectory(rootContext, "  ", toErase);
				} else if(command.compareToIgnoreCase("call")==0){
					ServerConfig server0011 = ServerConfig.getServerConfig(rootContext, "001", true);
					server0011.loader.loadProgram(server0011.serverLoadingCommand);
					ServerConfig monitor1 = ServerConfig.getMonitorConfig(rootContext,1);
					monitor1.loader.loadProgram(monitor1.monitorLoadingCommand);
				}
			} else if(args.length==0){
				toErase = false;
				System.out.println("<NameService>");
				ServerConfig.printNamingDirectory(rootContext, "  ", toErase);
			}
*/
		}
		catch (Exception e) {
			System.out.println("Error happened when using naming context");
		}
	}

	/**
	 * Constructor, initialize the object using default value
	 */
	public ServerConfig(){
		defaultValue();
	}
	
	/**
	 * Set the root naming context using the ORB object
	 * @param orb the ORB object
	 */
	public void setServerConfig(ORB orb){
		try{
			this.orb = orb;
			this.rootContext = NamingContextExtHelper.narrow(orb.resolve_initial_references("NameService"));
		}catch(Exception e){
			// retrieve root naming context failed, do nothing
			System.out.println("Root naming service access failed.");
		}
	}

	/**
	 * Set the root naming context
	 * @param rootContext The root naming context
	 */
	public void setServerConfig(NamingContextExt rootContext){
		defaultValue();
		this.rootContext = rootContext;
	}

	/**
	 * Set the object using default value
	 */
	public void defaultValue(){
		String IPs[] = ServerConfig.getLocalIP();
		if(serverConfigFileName.length()==0) centralNSConfigFileName = "centralNS.cfg";
		if(serverConfigFileName.length()==0) serverConfigFileName = "server.cfg";
		if(monitorConfigFileName.length()==0) monitorConfigFileName = "monitor.cfg";
		if(centralNCInitialHost.length()==0) centralNCInitialHost = IPs[IPs.length-1];
		if(centralNCInitialPort.length()==0) centralNCInitialPort = "3000";
		if(centralNCPort.length()==0) centralNCPort = "3002";
		if(localNCInitialPort.length()==0) localNCInitialPort = "3000";
		if(localNCPort.length()==0) localNCPort = "3002";
		if(localhostIP.length()==0) localhostIP = IPs[IPs.length-1];
		if(UDPPort.length()==0) UDPPort="3010";
		if(backupUDPPort.length()==0) backupUDPPort="3011";
		if(transferPort.length()==0) transferPort="3020";
	}

	/**
	 * @param isServer  Specify if it's going to read server config or monitor config 
	 * @throws Exception
	 */
	public void readConfigFromFile(boolean isReadingServerConfig) throws Exception{
		File configFile;	// used to read configuration

		// Read central naming service config
		configFile = new File(this.centralNSConfigFileName);
		// Set current working path
		String configPath = configFile.getAbsolutePath();
		this.currentDirectory = configPath.substring(0,configPath.lastIndexOf('\\'));
		// System.out.println("Working in " + currentDirectory);
		try{
			BufferedReader configReader = new BufferedReader(new FileReader(configFile));
			// Read and parse central naming service configuration
			try {
				while (true) {
					this.parseConfig(configReader.readLine());
				}
			}
			catch (Exception ex) {
				// Reach the end of file or other error happened
			}
			finally {
				configReader.close();
			}
		}catch(Exception e){
			// No central naming service config found, ignor it and let other config file do the job
		}
		
		
		// Read server or monitor config
		if(isReadingServerConfig){
			configFile = new File(this.serverConfigFileName);
			this.isServer = true;
		}else{
			configFile = new File(this.monitorConfigFileName);
			this.isMonitor = true;
		}

		BufferedReader configReader = new BufferedReader(new FileReader(configFile));
		// Read and parse server or monitor configuration
		try {
			while (true) {
				this.parseConfig(configReader.readLine());
			}
		}
		catch (Exception ex) {
			// Reach the end of file or other error happened
		}
		finally {
			configReader.close();
		}

		// Setup the loading command
		if (isReadingServerConfig) {
			if (this.serverLoadingCommand.length() == 0) {
				// this.serverLoadingCommand = "java -classpath " + currentDirectory + " BankServerLoader \"Server on "
				//		+ this.localhostIP + ":" + this.transferPort + "," + this.UDPPort + "\"";
				this.serverLoadingCommand = "java -classpath " + currentDirectory + " BankServerLoader " + currentDirectory;
			}
		}
		else {
			if (this.monitorLoadingCommand.length() == 0) {
				//this.monitorLoadingCommand = "java -classpath " + currentDirectory + " ServerConfig.tester \"Monitor on "
				//		+ this.localhostIP + ":" + this.transferPort + "," + this.UDPPort + "\"";
				this.monitorLoadingCommand = "java -classpath " + currentDirectory + " BankMonitor " + currentDirectory;
			}
		}

	}
	
	/**
	 * Read central naming service configure
	 * @throws Exception
	 */
	public void readNSFromFile() throws Exception{
		File configFile;	// used to read configuration

		// Read central naming service config
		configFile = new File(this.centralNSConfigFileName);
		// Set current working path
		String configPath = configFile.getAbsolutePath();
		this.currentDirectory = configPath.substring(0,configPath.lastIndexOf('\\'));
		// System.out.println("Working in " + currentDirectory);
		try{
			BufferedReader configReader = new BufferedReader(new FileReader(configFile));
			// Read and parse central naming service configuration
			try {
				while (true) {
					this.parseConfig(configReader.readLine());
				}
			}
			catch (Exception ex) {
				// Reach the end of file or other error happened
			}
			finally {
				configReader.close();
			}
		}catch(Exception e){
			// No central naming service config found, ignor it and let other config file do the job
		}
	}
	
	/**
	 * Parse the config item string and set the configuration accordingly
	 * @param configItem A string that have one configuration item (in the form of item:value)
	 */
	public void parseConfig(String configItem) {
		// Check if config information is in the String, if not, end the process.
		if (configItem.indexOf(':') == -1) {
			return;
		}

		// Extract the config and set the proporties of this object
		String configTitle = configItem.substring(0, configItem.indexOf(':')).trim();
		String configBody = configItem.substring(configItem.indexOf(':') + 1, configItem.length()).trim();

		if (configTitle.compareToIgnoreCase("centralNCInitialHost") == 0) {
			centralNCInitialHost = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("centralNCInitialPort") == 0) {
			centralNCInitialPort = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("centralNCPort") == 0) {
			centralNCPort = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("localNCInitialPort") == 0) {
			this.localNCInitialPort = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("localNCPort") == 0) {
			this.localNCPort = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("localhostIP") == 0) {
			this.localhostIP = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("branchID") == 0) {
			this.branchID = configBody;
			this.dataFileName = this.currentDirectory + "\\" + "branch" + configBody + ".dat";
			this.logFileName = this.currentDirectory + "\\" + "branch" + configBody + ".log";
			return;
		}
		if (configTitle.compareToIgnoreCase("monitorID") == 0) {
			this.monitorID = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("serverID") == 0) {
			this.serverID = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("dataFileName") == 0) {
			this.dataFileName = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("logFileName") == 0) {
			this.logFileName = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("transferPort") == 0) {
			this.transferPort = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("UDPPort") == 0) {
			this.UDPPort = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("backupUDPPort") == 0) {
			this.backupUDPPort = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("serverLoadingCommand") == 0) {
			this.serverLoadingCommand = configBody;
			return;
		}
		if (configTitle.compareToIgnoreCase("monitorLoadingCommand") == 0) {
			this.monitorLoadingCommand = configBody;
			return;
		}
		return;
	}

	/**
	 * Print the contents in a Naming context, include contents in subdirectory
	 * 
	 * @param currentContext  The context to start from
	 * @param space		The space to print before the content
	 * @param toErase   Specify if the structure will be cleared
	 */
	public static void printNamingDirectory(NamingContextExt currentContext, String space, boolean toErase) {

		// Build name list
		BindingListHolder bl = new BindingListHolder();
		BindingIteratorHolder blIt = new BindingIteratorHolder();
		currentContext.list(100, bl, blIt);
		Binding bindings[] = bl.value;

		// List contents in current directory
		for (int i = 0; i < bindings.length; i++) {

			int lastIx = bindings[i].binding_name.length - 1;

			// check to see if this is a naming context object
			if (bindings[i].binding_type == BindingType.ncontext) {
				System.out.println(space + "<" + bindings[i].binding_name[lastIx].id
						+ ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind))
						+ ">" + (toErase?" > to be erased":""));
				// List contents in the sub directory
				try {
					NamingContextExt subContext;
					subContext = NamingContextExtHelper.narrow(currentContext.resolve_str(bindings[i].binding_name[lastIx].id
									+ ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind))));
					printNamingDirectory(subContext, space + "  ", toErase);
					if (toErase) {
						try {
							NameComponent name[] = currentContext.to_name(bindings[i].binding_name[lastIx].id
									+ ((bindings[i].binding_name[lastIx].kind.length() == 0) ? ""
											: ("." + bindings[i].binding_name[lastIx].kind)));
							currentContext.unbind(name);
						}
						catch (Exception e) {
							System.out.println("Can not unbind "
									+ bindings[i].binding_name[lastIx].id
									+ ((bindings[i].binding_name[lastIx].kind.length() == 0) ? ""
											: ("." + bindings[i].binding_name[lastIx].kind)));
						}
					}
				}
				catch (Exception e) {
					System.out.println("Error happened when dig into subdirectory");
				}
			}
			else {
				// This is a normal object, print it directly
				System.out.println(space + "Obj:" + bindings[i].binding_name[lastIx].id
						+ ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind))
						+ (toErase?" > to be erased":""));
				if (toErase) {
					try {
						NameComponent name[] = currentContext.to_name(bindings[i].binding_name[lastIx].id
								+ ((bindings[i].binding_name[lastIx].kind.length() == 0) ? ""
										: ("." + bindings[i].binding_name[lastIx].kind)));
						currentContext.unbind(name);
					}
					catch (Exception e) {
						System.out.println("Can not unbind "
								+ bindings[i].binding_name[lastIx].id
								+ ((bindings[i].binding_name[lastIx].kind.length() == 0) ? ""
										: ("." + bindings[i].binding_name[lastIx].kind)));
					}
				}
			}
		}
	}

	/**
	 * Generate the naming context of the loader
	 * 
	 * @return return the nc string of the loader object
	 */
	public String getLoaderNC() {
		String nc = "";
		nc = "corbaname::" + this.localhostIP + ":" + this.localNCPort + "#loader" + this.branchID + "-"
				+ this.serverID;
		return nc;
	}
	
	
	
	/**
	 * Get the IP address of localhost
	 * @return return the  IP address list of localhost
	 */
	public static String[] getLocalIP() {
		InetAddress localAddresses[];
		String localIP[] = new String[0];
		try {
			java.net.InetAddress localMachine = java.net.InetAddress.getLocalHost();
			localAddresses = InetAddress.getAllByName(localMachine.getHostName());
			localIP = new String[localAddresses.length];
			for(int loop=0; loop<localAddresses.length;loop++){
				localIP[loop] = localAddresses[loop].getHostAddress();
			}
		}
		catch (java.net.UnknownHostException uhe) {
		}
		return localIP;
	}
	
	
	
	/**
	 * Get the server configuration by branchID and the primary tag
	 * @param rootContext The root naming context from where to extract server configuration
	 * @param branchID The branch ID of the server
	 * @param isPrimary To get the primary server or the backup server
	 * @return return the ServerConfig object which contain server config
	 */
	public static ServerConfig getServerConfig(NamingContextExt rootContext, String branchID, boolean isPrimary) throws Exception{
		ServerConfig theServer = new ServerConfig();
		theServer.setServerConfig(rootContext);
		theServer.isMonitor = false;
		theServer.isServer = true;
		theServer.branchID = branchID;
		
		// Find the naming context of the branch
		NamingContext ctx2 = NamingContextExtHelper.narrow(rootContext.resolve_str(ServerConfig.branchDirectory));
		ctx2 = NamingContextExtHelper.narrow(((NamingContextExt)ctx2).resolve_str(ServerConfig.branchTag+branchID));
		
		// Find who is the primary server
		
		// Build name list
		BindingListHolder bl = new BindingListHolder();
		BindingIteratorHolder blIt = new BindingIteratorHolder();
		ctx2.list(100, bl, blIt);
		Binding bindings[] = bl.value;
		// Search the string pattern "Primary:ServerX"
		String serverName = "";
		String serverID="";
		for (int i = 0; i < bindings.length; i++) {
			// Check each name and find primary server
			int lastIx = bindings[i].binding_name.length - 1;
			String currentName = bindings[i].binding_name[lastIx].id
			+ ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind));
			if(currentName.indexOf("Primary:")==-1){
				// Not the pattern, continue to next
				continue;
			}
			// Found the pattern "Primary:", extract the primary server name
			serverName = currentName.substring(currentName.indexOf(':')+1,currentName.length()).trim();
			break;
		}
		// Check if the primary server is found
		if(serverName.length()<7){
			// Not found, use Server1 as default
			serverName = "Server1";
		}
		if(!isPrimary){
			// Looking for the backup server
			if(serverName.compareToIgnoreCase("Server1")==0){
				serverName="Server2";
			}else if(serverName.compareToIgnoreCase("Server2")==0){
				serverName="Server1";
			}
		}
		serverID = serverName.substring(6,7);
		ctx2 = NamingContextExtHelper.narrow(((NamingContextExt)ctx2).resolve_str(serverName));
		
		// Retrieve server information
		ctx2.list(100, bl, blIt);
		bindings = bl.value;
		for (int i = 0; i < bindings.length; i++) {
			// Retrieve configuration
			int lastIx = bindings[i].binding_name.length - 1;
			theServer.parseConfig(bindings[i].binding_name[lastIx].id
					+ ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind)));
		}
		org.omg.CORBA.Object obj = ((NamingContextExt)ctx2).resolve_str(ServerConfig.loaderTag
				+ ServerConfig.branchTag
				+ branchID + "-" + serverID);
		theServer.loader = ProgramLoaderHelper.narrow(obj);
		
		return theServer;
	}



	/**
	 * Get the server configuration by branchID and serverID
	 * @param rootContext The root naming context from where to extract server configuration
	 * @param branchID The branch ID of the server
	 * @param serverID To get the server1 or the server2
	 * @return return the ServerConfig object which contain server config
	 */
	public static ServerConfig getServerConfig(NamingContextExt rootContext, String branchID, String serverID) throws Exception{
		ServerConfig theServer = new ServerConfig();
		theServer.setServerConfig(rootContext);
		theServer.isMonitor = false;
		theServer.isServer = true;
		theServer.branchID = branchID;
		theServer.serverID = serverID;
		
		// Find the naming context of the branch
		NamingContext ctx2 = NamingContextExtHelper.narrow(rootContext.resolve_str(ServerConfig.branchDirectory));
		ctx2 = NamingContextExtHelper.narrow(((NamingContextExt)ctx2).resolve_str(ServerConfig.branchTag+branchID));
		ctx2 = NamingContextExtHelper.narrow(((NamingContextExt)ctx2).resolve_str(ServerConfig.serverTag+serverID));
		
		// Build name list
		BindingListHolder bl = new BindingListHolder();
		BindingIteratorHolder blIt = new BindingIteratorHolder();
		// Retrieve server information
		ctx2.list(100, bl, blIt);
		Binding bindings[] = bl.value;
		
		bindings = bl.value;
		for (int i = 0; i < bindings.length; i++) {
			// Retrieve configuration
			int lastIx = bindings[i].binding_name.length - 1;
			theServer.parseConfig(bindings[i].binding_name[lastIx].id
					+ ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind)));
		}
		org.omg.CORBA.Object obj = ((NamingContextExt)ctx2).resolve_str(ServerConfig.loaderTag
				+ ServerConfig.branchTag
				+ branchID + "-" + serverID);
		theServer.loader = ProgramLoaderHelper.narrow(obj);
		
		return theServer;
	}

	
	/**
	 * Get the monitor configuration
	 * @param rootContext The root naming context from where to extract server configuration
	 * @param monitorID The monitor ID of the server
	 * @return return the ServerConfig object which contain server config
	 */
	public static ServerConfig getMonitorConfig(NamingContextExt rootContext, String monitorID) throws Exception{
		ServerConfig theServer = new ServerConfig();
		theServer.setServerConfig(rootContext);
		theServer.isMonitor = true;
		theServer.isServer = false;
		theServer.monitorID = monitorID;
		
		// Find the naming context of the branch
		NamingContext ctx2 = NamingContextExtHelper.narrow(rootContext.resolve_str(ServerConfig.monitorDirectory));
		ctx2 = NamingContextExtHelper.narrow(((NamingContextExt)ctx2).resolve_str(ServerConfig.monitorTag + monitorID));
		
		// Build name list
		BindingListHolder bl = new BindingListHolder();
		BindingIteratorHolder blIt = new BindingIteratorHolder();
		ctx2.list(100, bl, blIt);
		Binding bindings[] = bl.value;

		for (int i = 0; i < bindings.length; i++) {
			// Retrieve configuration
			int lastIx = bindings[i].binding_name.length - 1;
			theServer.parseConfig(bindings[i].binding_name[lastIx].id
					+ ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind)));
		}
		org.omg.CORBA.Object obj = ((NamingContextExt)ctx2).resolve_str(ServerConfig.loaderTag
				+ ServerConfig.monitorTag
				+ monitorID);
		theServer.loader = ProgramLoaderHelper.narrow(obj);
		
		return theServer;
	}
	
	/**
	 * Remote launch a server or a monitor
	 * @throws Exception
	 */
	public void relaunch() throws Exception{
		if(this.isServer){
			// Relaunch a server
			this.loader.loadProgram(this.serverLoadingCommand);
		}
		if(this.isMonitor){
			// Relaunch a server
			this.loader.loadProgram(this.monitorLoadingCommand);
		}
	}

	/**
	 * Setup the server information in the naming service
	 * Branch ID and server ID are required to build the structure in naming context
	 * @param rootContext The root context of naming service
	 * @throws Exception Throw Exception when error happened
	 */
	public void setupServer(NamingContextExt rootContext) throws Exception{
		if(!this.isServer){
			throw new Exception("This is not a server configure.");
		}

		// Add branches directory if it does not exist
		NameComponent name[] = rootContext.to_name(ServerConfig.branchDirectory);
		NamingContext ctx2 = rootContext;
		try {
			ctx2 = ctx2.bind_new_context(name);
		} catch (Exception er) {
			if (er instanceof AlreadyBound) {
				// Already bound, get the context directly
				ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.branchDirectory));
			}
		}

		// Add specific branch if it does not exist
		name = rootContext.to_name(ServerConfig.branchTag + this.branchID);
		try {
			ctx2 = ctx2.bind_new_context(name);
		} catch (Exception er) {
			if (er instanceof AlreadyBound) {
				// Already bound, get the context directly
				ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.branchTag + this.branchID));
			}
		}

		// Add specific server if it does not exist
		name = rootContext.to_name(ServerConfig.serverTag + this.serverID);
		try {
			ctx2 = ctx2.bind_new_context(name);
		} catch (Exception er) {
			if (er instanceof AlreadyBound) {
				// Already bound, get the context directly
				ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.serverTag + this.serverID));
			}
		}
		
		// Erase the content in current directory
		ServerConfig.printNamingDirectory(NamingContextExtHelper.narrow(ctx2), "", true);
		
		// Add server information under the server directory
		name = rootContext.to_name("localhostIP: " + this.localhostIP);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name("UDPPort: " + this.UDPPort);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name("backupUDPPort: " + this.backupUDPPort);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name("transferPort: " + this.transferPort);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name("serverLoadingCommand: " + this.serverLoadingCommand);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name(ServerConfig.loaderTag + ServerConfig.branchTag + this.branchID + "-"
				+ this.serverID);
		ctx2.rebind(name, this.loader);
	}
	
	/**
	 * Setup the monitor information in the naming service
	 * Monitor ID is required to build the structure in naming context
	 * @param rootContext The root context of naming service
	 * @throws Exception Throw Exception when error happened
	 */
	public void setupMonitor(NamingContextExt rootContext) throws Exception{
		if(!this.isMonitor){
			throw new Exception("This is not a monitor configure.");
		}

		// Add monitors directory if it does not exist
		NameComponent name[] = rootContext.to_name(ServerConfig.monitorDirectory);
		NamingContext ctx2 = rootContext;
		try {
			ctx2 = ctx2.bind_new_context(name);
		} catch (Exception er) {
			if (er instanceof AlreadyBound) {
				// Already bound, get the context directly
				ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.monitorDirectory));
			}
		}

		// Add specific mointor if it does not exist
		name = rootContext.to_name(ServerConfig.monitorTag + this.monitorID);
		try {
			ctx2 = ctx2.bind_new_context(name);
		} catch (Exception er) {
			if (er instanceof AlreadyBound) {
				// Already bound, get the context directly
				ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.monitorTag + this.monitorID));
			}
		}

		// Erase the content in current directory
		ServerConfig.printNamingDirectory(NamingContextExtHelper.narrow(ctx2), "", true);

		// Add server information under the server directory
		name = rootContext.to_name("localhostIP: " + this.localhostIP);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name("UDPPort: " + this.UDPPort);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name("backupUDPPort: " + this.backupUDPPort);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name("monitorLoadingCommand: " + this.monitorLoadingCommand);
		ctx2.rebind(name, ctx2);
		name = rootContext.to_name(ServerConfig.loaderTag + ServerConfig.monitorTag + this.monitorID);
		ctx2.rebind(name, this.loader);
	}
	
	/**
	 * List current monitor names in the naming service
	 * @param rootContext The root context of naming service
	 * @throws Exception Throw Exception when error happened
	 * @return Return a String array that contain the monitor list
	 */
	public static String[] listMonitors(NamingContextExt rootContext) throws Exception{
		String monitors[]=new String[0];

		NamingContext ctx2 = rootContext;
		// Get into monitors directory
		ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.monitorDirectory));
		
		// Build name list
		BindingListHolder bl = new BindingListHolder();
		BindingIteratorHolder blIt = new BindingIteratorHolder();
		ctx2.list(100, bl, blIt);
		Binding bindings[] = bl.value;
		monitors = new String[bindings.length];

		for (int i = 0; i < bindings.length; i++) {
			// List monitor names in the directory
			int lastIx = bindings[i].binding_name.length - 1;
			String item = bindings[i].binding_name[lastIx].id + ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind));
			// Check if the content is a monitor configure
			if(bindings[i].binding_type == BindingType.ncontext && item.length()>=8){
				if(item.substring(0,7).compareToIgnoreCase(ServerConfig.monitorTag)==0){
					monitors[i]=item;
				}else{
					monitors[i]="";
				}
			}else{
				monitors[i]="";
			}
		}
		
		// Shrink the array to eliminate zero length strings
		monitors = ServerConfig.shrinkArray(monitors);

		return monitors;
	}

	/**
	 * List current server names in the naming service
	 * @param rootContext The root context of naming service
	 * @throws Exception Throw Exception when error happened
	 * @return Return a String array that contain the server list
	 */
	public static String[] listServers(NamingContextExt rootContext) throws Exception{
		String servers[]=new String[0];

		NamingContext ctx2 = rootContext;
		// Get into monitors directory
		ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.branchDirectory));
		
		// Build name list
		BindingListHolder bl = new BindingListHolder();
		BindingIteratorHolder blIt = new BindingIteratorHolder();
		ctx2.list(100, bl, blIt);
		Binding bindings[] = bl.value;
		servers = new String[bindings.length*2];

		for (int i = 0; i < bindings.length; i++) {
			// Browse the branch directory
			int lastIx = bindings[i].binding_name.length - 1;
			String branch = bindings[i].binding_name[lastIx].id + ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind));
			// Check if server1 in current branch exist
			try{
				NamingContextExt nc = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(branch));
				nc = NamingContextExtHelper.narrow(((NamingContextExt) nc).resolve_str("Server1"));
				// Server1 exists
				servers[i*2]=branch + "\\Server1";
			}catch(Exception e){
				// Server1 does not exist
				servers[i*2]="";
			}
			
			// Check if server2 in current branch exist
			try{
				NamingContextExt nc = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(branch));
				nc = NamingContextExtHelper.narrow(((NamingContextExt) nc).resolve_str("Server2"));
				// Server2 exists
				servers[i*2+1]=branch + "\\Server2";
			}catch(Exception e){
				// Server2 does not exist
				servers[i*2+1]="";
			}
		}

		// Shrink the array to eliminate zero length strings
		servers = ServerConfig.shrinkArray(servers);
		
		return servers;
	}
	
	/**
	 * @param oldarray The array to be shrinked
	 * @return Return the new array that does not contain null or zero length strings
	 */
	public static String[] shrinkArray(String[] oldArray){
		String[] newArray = new String[0];
		int concreteLength = 0;
		for(int i=0;i<oldArray.length;i++){
			// If the string is not null and not a zero length string, add the number
			if(oldArray[i]!=null){
				if(oldArray[i].length()>0){
					concreteLength++;
				}
			}
		}
		newArray = new String[concreteLength];
		
		// Copy the non-zero length strings to the new array
		int currentPosition = 0; // The current position in the new array
		for(int i=0;i<oldArray.length;i++){
			// If the string is not null and not a zero length string, copy it to new array
			if(oldArray[i]!=null){
				if(oldArray[i].length()>0){
					newArray[currentPosition++]=oldArray[i];
				}
			}
		}
	
		return newArray;
	}
	
	/**
	 * Notify a monitor to load a branch server
	 * @param rootContext 
	 * @param branchID
	 * @param isPrimary
	 */
	public static void notifyMonitor(NamingContextExt rootContext, String branchID, boolean isPrimary){
		// Generate a new thread to notify the deamon to launch the backup server of a branch
		ServerNotifier notifier = new ServerNotifier();
		notifier.branchID = branchID;
		notifier.rootContext = rootContext;
		notifier.setDaemon(true);
		notifier.start();
	}
	

	/**
	 * Get a unique ID from the central naming service
	 * @param rootContext The central naming service
	 * @return	Return the unique ID string
	 * @throws Exception  Throw exception when have problem contact with central naming service
	 */
	public static String getUniqueID(NamingContextExt rootContext) throws Exception{
		String uniqueID = "";
		long currentID = 0;
		boolean succeeded = false;
		
		// retrieve Unique ID from central naming service
		NamingContext ctx2 = rootContext;
		
		// Repeat untill get a Unique ID from central naming service
		while (!succeeded) {
			// Set the mutual exclution tag
			NameComponent name[] = rootContext.to_name(ServerConfig.uniqueTag);
			ctx2.rebind(name, ctx2);

			// List names in the root directory of the central naming service
			BindingListHolder bl = new BindingListHolder();
			BindingIteratorHolder blIt = new BindingIteratorHolder();
			ctx2.list(100, bl, blIt);
			Binding bindings[] = bl.value;

			String currentName;	// The current unique ID string, in the form of "Unique: xxxx"
			// Retrieve current unique ID
			for (int i = 0; i < bindings.length; i++) {
				int lastIx = bindings[i].binding_name.length - 1;
				currentName = bindings[i].binding_name[lastIx].id + ((bindings[i].binding_name[lastIx].kind.length()==0)?"":("."+bindings[i].binding_name[lastIx].kind));
				if(currentName.length()>(ServerConfig.uniqueTag+":").length()){
					// Find the string start with uniqueTag+":", get the current ID and erase it from naming context
					if(currentName.substring(0,(ServerConfig.uniqueTag+":").length()).compareToIgnoreCase(ServerConfig.uniqueTag+":")==0){
						// Get current ID
						try{
							currentID = Long.parseLong(currentName.substring(currentName.indexOf(":")+1, currentName.length()).trim());
						}catch(Exception e){
							// Invalid ID, do nothing
						}
						
						// Erase this ID from central naming service 
						try{
							name = rootContext.to_name(currentName);
							ctx2.unbind(name);
						}catch(Exception e){
							// Not exist or has been erased by others, do nothing
						}
					}
				}
			}
			
			// Set new unique ID
			currentID++;
			// Rewind the ID if it exceed the range
			if(currentID>Long.MAX_VALUE-100) currentID=0;
			
			name = rootContext.to_name(ServerConfig.uniqueTag + ":" + currentID);
			ctx2.rebind(name, ctx2);
			
			// Release tag and check if the unique ID has been taken by others
			try {
				name = rootContext.to_name(ServerConfig.uniqueTag);
				ctx2.unbind(name);
				succeeded = true;
			}
			catch (NotFound e) {
				// Indicate that the current Unique ID has been taken by others, need to get another unique ID
				succeeded = false;
			}
		}
		
		// Set the mutual exclusion tag in central naming service

		
		uniqueID = String.format("%1$" + BankUtility.OpIDPrefixLength + "d",currentID); 
		return uniqueID;
	}
}


/**
 * @author Min Huang (khtwo)
 * Thread to notify the deamon to launch the backup server of a branch
 */
class ServerNotifier extends Thread{
	public String branchID;
	public NamingContextExt rootContext;
	
	public void run(){
		Packet request = new Packet();
		request.status = Packet.RELOAD;
		request.opID = branchID;
		UDPClientSocket sender=new UDPClientSocket();
		
		// Find a running monitor and request it to relaunch the server
		try{
			String[] monitorList = ServerConfig.listMonitors(rootContext);
			for(int i=0; i<monitorList.length; i++){
				try{
					ServerConfig monitor = ServerConfig.getMonitorConfig(rootContext, monitorList[i].substring(ServerConfig.monitorTag.length(), ServerConfig.monitorTag.length()+1));
					sender.sendPacket(monitor.localhostIP,  Integer.parseInt(monitor.UDPPort), request);
					// Sending request to monitor succeeded, quit this function
					break;
				}catch(Exception e){
					// Sending request to monitor failed, continue
				}
			}
		}catch(Exception e){
			// Error while retrieving the monitor list, do nothing
		}
	}
}