import java.net.*;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.rmi.RemoteException;
import java.util.Properties;

import org.omg.CORBA.ORB;
import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContext;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAHelper;

import BankApp.*;
import BankServer.*;
import ServerConfig.ServerConfig;

/**
 * BankServerLoader is used to load the Bank Server. First, it will check the match between log file and ata file, then
 * it will test if current server can be run as a primary server. If so, a primary server will be launched, else a
 * backup server will be launched. When it become a backup server, it will detect if primary running correctly. When
 * primary server is found failed, the backup will try to upgrade to a primary server and call the monitor to load a
 * backup server.
 * 
 * @author Min Huang (khtwo)
 */
public class BankServerLoader {
	public static ServerConfig currentConfig = new ServerConfig(); // Current configure of this server

	public static UDPServerSocket UDPServer = null; // The socket for primary server to listen request

	public static UDPServerSocket backupUDPServer = null; // The socket for backup server to listen request

	public static ServerSocket TransferServer = null; // The socket to accept transfer

	public static ORB orb; // The orb

	public static BankImpl bankImpl = null; // The Primary Bank Server object reference

	public static boolean backupStarted = false;

	public static boolean backupRunning = true;

	public static boolean primaryFinished = false;

	public static boolean programEnd = false;

	public static String[] startArgs = null;

	public static int semaphoreTried = 0; // Times of tries has been made to get the semaphore

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// Add a hook to handle exit
		Runtime.getRuntime().addShutdownHook(new Thread() {
			public void run() {
				// System.out.println("System shutting down.");
				backupRunning = false;
				primaryFinished = true;
				programEnd = true;

				// Shutdown the primary Bank Server
				try {
					if (bankImpl != null)
						bankImpl.finalize();
				}
				catch (Throwable e) {
				}

				// Close the UDP server socket if it remains open
				if (UDPServer != null)
					// if (!UDPServer.isClosed())
					try{
						UDPServer.close();
					}catch(Exception e){}

				// Close the UDP server socket if it remains open
				if (backupUDPServer != null)
					// if (!UDPServer.isClosed())
					try{
						backupUDPServer.close();
					}catch(Exception e){}

				// Close the Transfer Server socket if it remains open
				if (TransferServer != null) {
					if (!TransferServer.isClosed())
						try {
							TransferServer.close();
						}
						catch (IOException e) {
						}
				}

				// debug
				if (BankUtility.DEBUG_LEVEL > 1) {
					try {
						System.out.println("Press any key to end");
						System.in.read();
					}
					catch (Exception e) {
					}
				}
			}
		});

		// Using specific working directory
		if (args.length > 0) {
			currentConfig.centralNSConfigFileName = args[0] + "\\" + currentConfig.centralNSConfigFileName;
			currentConfig.serverConfigFileName = args[0] + "\\" + currentConfig.serverConfigFileName;
			currentConfig.monitorConfigFileName = args[0] + "\\" + currentConfig.monitorConfigFileName;
			currentConfig.currentDirectory = args[0];
		}
		// Load configure from centralNS.cfg, server.cfg
		loadConfig();
		// Register the server on central naming service
		registerServer();

		System.out.println("Starting Bank Server" + currentConfig.serverID + " in branch " + currentConfig.branchID);

		// Check the server type and run in corresponding mode
		runServer();

	}

	
	
	/**
	 * Load configure from centralNS.cfg, server.cfg
	 */
	public static void loadConfig() {
		// Try to read server config
		try {
			currentConfig.readConfigFromFile(true);
		}
		catch (Exception e) {
			// No config is read, do nothing and quit
			System.out
					.println("Reading configure failed. \nPlease check if centralNS.cfg, server.cfg or monitor.cfg exists.");
			Runtime.getRuntime().exit(1);
		}

		// Retrieve the central naming service
		Properties props = new Properties();
		props.put("org.omg.CORBA.ORBInitialPort", ServerConfig.centralNCInitialPort);
		props.put("org.omg.CORBA.ORBInitialHost", ServerConfig.centralNCInitialHost);
		String empty[] = new String[0];
		orb = ORB.init(empty, props);
		currentConfig.setServerConfig(orb);
	} // End of loadConfig

	/**
	 * Register the server on central naming service
	 */
	public static void registerServer() {
		// Check if this server has been registered on central naming service
		try {
			ServerConfig.getServerConfig(currentConfig.rootContext, currentConfig.branchID, currentConfig.serverID);
		}
		catch (Exception e) {
			// This server has not been registered, register it on central naming service
			try {
				System.out
						.println("The server information is not found on central naming service.\nRegistering the server...  ");
				Runtime.getRuntime().exec("java ServerConfig.RegisterProgramLoader").waitFor();
				System.out.println("Server registration OK.");
			}
			catch (Exception e1) {
				// Registration failed, quit
				System.out.println("Server registration failed.");
				Runtime.getRuntime().exit(1);
			}
		}
	} // End of registerServer

	
	
	/**
	 * Check the server type and run in corresponding mode
	 */
	public static void runServer() {
		boolean isDirectPrimary = false;
		NamingContext ctx2 = currentConfig.rootContext;
		NameComponent name[] = new NameComponent[0];

		// Check if the primary server is running
		try {
			ctx2 = NamingContextExtHelper.narrow(currentConfig.rootContext.resolve_str(ServerConfig.branchDirectory));
			ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.branchTag
					+ currentConfig.branchID));
			name = currentConfig.rootContext.to_name(ServerConfig.PrimaryTag);
			ctx2.bind(name, ctx2);
			System.out.println("Primary Server not found, running as primary server.");
			isDirectPrimary = true;
		}
		catch (Exception e) {
			// Binding failed, there is a primary server running, run as a backup server
			System.out.println("Primary Server found, running as a backup server.");
			isDirectPrimary = false;
		}

		try {
			// If primary server has not started, continue running backup server
			while (!primaryFinished && !programEnd) {
				// Check current status
				if (!isDirectPrimary) {
					// Run as backup server
					runBackup();
				}
				// Switch to primary server
				if (!programEnd)
					runPrimary(isDirectPrimary);
			}
		}
		catch (Exception e) {
			// Error happened in the running, normally another instance of server is running
			// System.out.println(e);
		}

	}// End of runServer

	
	
	/**
	 * Run as a backup server
	 * 
	 * @throws Exception
	 *             Throw when the socket port has been taken or another instance of server is running
	 */
	public static void runBackup() throws Exception {
		// Initialize the UDP listener
		if (!BankServerLoader.backupStarted) {
			try {
				// Start up the backup UDP server
				BankServerLoader.backupUDPServer = new UDPServerSocket(Integer.parseInt(currentConfig.backupUDPPort));
			}
			catch (NumberFormatException e) {
				System.out.println("Bad configure found in current configure.");
				System.exit(1);
			}
			catch (Exception r) {
				System.out.println("Can not bind to the port, maybe another instance of server is running");
				Runtime.getRuntime().exit(1);
			}
		}

		// Ask the primary server for a log synchronization
		UDPClientSocket sender = new UDPClientSocket();
		ServerConfig primaryConfig = null;
		LogManager log = new LogManager(currentConfig.logFileName);
		RandomAccessFile dataFile = new RandomAccessFile(currentConfig.dataFileName, "rw");
		// Check if the datafile exists
		if (dataFile.length() == 0)// this is the new file
		{
			// Create the new file and set to proper length
			dataFile.setLength(BankUtility.ACCOUNT_SIZE * 10000);
			dataFile.seek(0);// starting
			for (int j = 0; j < 10000; j++) {
				// branchFiles[i].writeInt(0);//initialize to 0;
				dataFile.writeFloat(BankUtility.NegativeInfinite);// indicating unused
			}
		}

		// Check log file and data file
		if (!BankServerLoader.backupStarted) {
			BankServerLoader.checkLogFile(log, dataFile);
		}

		// Retrieve primary server information
		try {
			String otherServerID = currentConfig.serverID.compareTo("1") == 0 ? "2" : "1";
			primaryConfig = ServerConfig.getServerConfig(currentConfig.rootContext, currentConfig.branchID,
					otherServerID);
		}
		catch (Exception e) {
			// Can not find primary server information, return and upgrade to primary server
			return;
		}

		// Notify the primary server that the backup server is joining
		Packet joinPacket = new Packet();
		joinPacket.status = BankUtility.REQUEST_JOIN;
		try {
			sender.sendPacket(primaryConfig.localhostIP, Integer.parseInt(primaryConfig.UDPPort), joinPacket);
			semaphoreTried = 0;
		}
		catch (Exception e) {
			// Can not contact with primary server, return and upgrade to primary server
			return;
		}


		int selfChecking = 0;  // Check if the server itself is still in the central naming service
		// Wait for log messages from primary server
		boolean isJoinFinished = false; // If the join operation is finished, find by time out
		while (backupRunning) {
			try {
				// Receive log message
				Packet receivedLog = BankServerLoader.backupUDPServer.recvPacket();

				semaphoreTried = 0;

				// Ignore test packet
				if (receivedLog.status == Packet.TEST || receivedLog.status == Packet.TEST_REPLY) {
					continue;
				}

				BankUtility.showDebugInfo(1, "Received a log Packet");

				processLog(log, receivedLog, dataFile);

			}
			catch (Exception e) {
				// Check if the server itself is still in the central naming service
				if(selfChecking++ > 5){
					selfChecking=0;
					try{
						ServerConfig myself = ServerConfig.getServerConfig(currentConfig.rootContext, currentConfig.branchID, currentConfig.serverID);
					}catch(Exception r){
						// The server registration is not in the central naming service, shutdown
						System.out.println("Server registration not found, quit");
						System.exit(1);
					}
				}
				
				
				// Long time that not receive any log from server
				// Check if the primary server is ok, if not, this backup server upgrade to primary server
				isJoinFinished = true;

				Packet testPacket = new Packet();
				testPacket.status = Packet.TEST;
				try {
					if (!programEnd)
						sender.sendPacket(primaryConfig.localhostIP, Integer.parseInt(primaryConfig.UDPPort),
								testPacket);
					semaphoreTried = 0;
				}
				catch (Exception e1) {
					// Can not contact with primary server, stop the backup server and upgrade to primary server
					System.out.println("Contact with primary server failed.");
					backupRunning = false;
				}
			}
		}
		
		dataFile.close();

		BankServerLoader.backupStarted = true;

	} // End of runBackup

	/**
	 * Run as a primary server
	 * 
	 * @param isDirectPrimary
	 *            True if this execution is a direct Primary server execution, false if it's an upgrade from a backup
	 *            server
	 * @throws Exception
	 *             Throw when the socket port has been taken or another instance of server is running
	 */
	public static void runPrimary(boolean isDirectPrimary) throws Exception {
		LogManager log = new LogManager(currentConfig.logFileName);
		RandomAccessFile dataFile = new RandomAccessFile(currentConfig.dataFileName, "rw");
		// Check if the datafile exists
		if (dataFile.length() == 0)// this is the new file
		{
			// Create the new file and set to proper length
			dataFile.setLength(BankUtility.ACCOUNT_SIZE * 10000);
			dataFile.seek(0);// starting
			for (int j = 0; j < 10000; j++) {
				// branchFiles[i].writeInt(0);//initialize to 0;
				dataFile.writeFloat(BankUtility.NegativeInfinite);// indicating unused
			}
		}

		// Check log file and data file
		if (isDirectPrimary) {
			BankServerLoader.checkLogFile(log, dataFile);
		}

		// Try to register as a primary server, the naming context "BranchXXX" is used as the semaphore
		NameComponent semaphore[] = currentConfig.rootContext.to_name(ServerConfig.PrimaryTag + currentConfig.branchID);
		try {
			currentConfig.rootContext.bind(semaphore, currentConfig.rootContext);
		}
		catch (Exception e) {
			// The semaphore has been taken by others, test if the semaphore is permanently locked
			semaphoreTried++;
			if (semaphoreTried >= ServerConfig.semaphoreRetry) {
				// Too many times retry, the semaphore might be permanently locked, release it
				currentConfig.rootContext.unbind(semaphore);
			}
			return;
		}

		// get reference to rootpoa & activate the POAManager
		POA rootpoa = POAHelper.narrow(orb.resolve_initial_references("RootPOA"));
		rootpoa.the_POAManager().activate();

		// create servant and register it with the ORB
		try {
			bankImpl = new BankImpl(currentConfig);
		}
		catch (Exception e) {
			System.out.println("Can not bind to the port, maybe another instance is running");
			Runtime.getRuntime().exit(1);
		}

		BankServerLoader.UDPServer = bankImpl.UDPServer;
		BankServerLoader.TransferServer = bankImpl.transferServer;

		// Check if the starting primary Bank Server succeeded
		bankImpl.setORB(orb);

		// get object reference from the servant
		org.omg.CORBA.Object ref = rootpoa.servant_to_reference(bankImpl);
		// and cast the reference to a CORBA reference
		Bank href = BankHelper.narrow(ref);

		// Search for the current branch naming context
		NamingContext ctx2 = currentConfig.rootContext;
		ctx2 = NamingContextExtHelper.narrow(currentConfig.rootContext.resolve_str(ServerConfig.branchDirectory));
		ctx2 = NamingContextExtHelper.narrow(((NamingContextExt) ctx2).resolve_str(ServerConfig.branchTag
				+ currentConfig.branchID));
		// bind the "Primary" server object Reference in current branch naming context
		String name = ServerConfig.PrimaryTag;
		NameComponent path[] = currentConfig.rootContext.to_name(name);
		ctx2.rebind(path, href);

		// Rebind Primary Server Distinguisher
		try {
			// Unbind the name that direct the primary server to another Bank Server
			path = currentConfig.rootContext.to_name(ServerConfig.PrimaryTag + ":Server"
					+ (currentConfig.serverID.compareTo("1") == 0 ? "2" : "1"));
			ctx2.unbind(path);
		}
		catch (Exception e) {/* Do nothing */
		}
		// Bind new Primary Server Distinguisher
		path = currentConfig.rootContext.to_name(ServerConfig.PrimaryTag + ":Server" + currentConfig.serverID);
		ctx2.rebind(path, ctx2);

		// If this is an upgrad from backup server, call the monitor to reload another Bank Server
		if (!isDirectPrimary) {
			System.out.println("Upgrading from backup server to primary server");

			// Depreciated, call the loader directly
			// ServerConfig anotherServer = ServerConfig.getServerConfig(currentConfig.rootContext,
			// currentConfig.branchID,(currentConfig.serverID.compareTo("1") == 0?"2":"1"));
			// anotherServer.relaunch();
		}
		else {
			// System.out.println("Running as primary server");
			// Initialize the UDP listener
			try {
				// Start up the backup UDP server
				BankServerLoader.backupUDPServer = new UDPServerSocket(Integer.parseInt(currentConfig.backupUDPPort));
			}
			catch (NumberFormatException e) {
				System.out.println("Bad configure found in current configure.");
				System.exit(1);
			}
			catch (Exception r) {
				System.out.println("Can not bind to the port, maybe another instance of server is running");
				Runtime.getRuntime().exit(1);
			}
		}

		// Release the binding semaphore after 2 seconds
		Thread releaseSemaphore = new Thread() {
			public NameComponent semaphore[] = currentConfig.rootContext.to_name(ServerConfig.PrimaryTag
					+ currentConfig.branchID);

			public void run() {
				try {
					Thread.sleep(1000);
					currentConfig.rootContext.unbind(semaphore);
				}
				catch (Exception e) {
				}
			}
		};
		releaseSemaphore.start();

		// Notify the monitor to start backup server
		System.out.println("Request the monitor to relaunch backup server of branch" + currentConfig.branchID);
		ServerConfig.notifyMonitor(currentConfig.rootContext, currentConfig.branchID, false);

		// Launch primary server and wait for call
		orb.run();

		/*
		 * Depreciated // Wait the primary server to die while(bankImpl.running){ try{ Thread.sleep(1000);
		 * }catch(Exception e){ // Clean the running resource try { bankImpl.finalize(); } catch (Throwable e1) { } } }
		 */

	} // End of runPrimary

	/**
	 * Process the received log packet
	 * 
	 * @param log
	 *            The Log Manager
	 * @param logPacket
	 *            Log Packet received
	 * @param dataFile
	 *            The data file to write
	 */
	public static void processLog(LogManager log, Packet logPacket, RandomAccessFile dataFile) {
		// Check if the log has been processed
		if (log.findRecord(logPacket) != -1) {
			// Already processed, do nothing
			return;
		}

		try {
			// Process the log packet
			int localAcnt = logPacket.source % 10000;
			dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
			float oldValue = dataFile.readFloat();
			// Check if the data match the log
			if (oldValue == logPacket.oldValue) {
				// Write new value to the account
				dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
				dataFile.writeFloat(logPacket.newValue);
			}
			else {
				// The log sequence might be messed up, add up the difference
				dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
				dataFile.writeFloat(oldValue + (logPacket.newValue - logPacket.oldValue));
			}

			// Record this log packet
			log.writePacket(logPacket);
			log.setCheckPoint();
		}
		catch (Exception e) {
			// Error happened in processing the log packet
		}
	}

	/**
	 * Check if the log match the data file, if not, correct it. This is done when the first time the program restart.
	 * 
	 * @param log
	 *            The Log Manager
	 * @param dataFile
	 *            The data file to write
	 */

	public static void checkLogFile(LogManager log, RandomAccessFile dataFile) {
		// Only check the last log, because the sequence of writting log and data operation are
		// fixed.
		if (log.getCurrentPos() > 0) {
			log.setCurrentPos(log.getCurrentPos() - 1);
			Packet logPacket = null;
			try {
				logPacket = log.readPacket();
			}
			catch (RemoteException e) {
				// Log reading failed, quit
				Runtime.getRuntime().exit(1);
			}

			if (logPacket == null) {
				// Log reading failed, quit
				Runtime.getRuntime().exit(1);
			}

			try {
				if (logPacket.opType != BankUtility.TRANSFER) {
					// If the operation is a non transfer operation
					int localAcnt = logPacket.source % 10000;
					dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
					float oldValue = dataFile.readFloat();
					// Check if the data match the log
					if (oldValue == logPacket.oldValue) {
						// Write new value to the account
						dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
						dataFile.writeFloat(logPacket.newValue);
					}
				}
				else {
					// The operation is a transfer operation
					if (logPacket.source / 10000 == logPacket.dest / 10000) {
						// It's a local transfer, Check the value
						int localAcnt = logPacket.source % 10000;
						dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
						float oldValue = dataFile.readFloat();
						// Check if the data match the log
						if (oldValue == logPacket.oldValue) {
							// Write new value to the source account
							dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
							dataFile.writeFloat(logPacket.newValue);
							// Write new value to the destination account
							localAcnt = logPacket.dest % 10000;
							dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
							oldValue = dataFile.readFloat();
							dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
							dataFile.writeFloat(oldValue + logPacket.newValue - logPacket.oldValue);
						}
					}
					else {
						// It's a remote transfer, redo the transfer
						// Setup the packet for transfer operation
						Packet receivePacket = new Packet();
						Packet transferPacket = new Packet();
						transferPacket.opID = logPacket.opID;
						transferPacket.opType = BankUtility.TRANSFER;
						transferPacket.source = logPacket.source;
						transferPacket.dest = logPacket.dest;
						transferPacket.oldValue = logPacket.newValue - logPacket.oldValue;// this is no good;
						transferPacket.newValue = -1;// should be ignored

						boolean done = false;
						int retry = 0;

						// Redo the transfer, retry 5 times when failed
						while (!done && retry < 5) {
							retry++;
							try {
								String remoteBranch = "000";
								String remoteHostName = "", remotePortNumbers = "";
								String tempString = String.format("%1$3d", logPacket.dest / 10000).trim();

								if (tempString.length() < 3) {
									remoteBranch = remoteBranch.subSequence(0, 3 - tempString.length()) + tempString;
								}
								else {
									remoteBranch = tempString;
								}

								// Get the remote branch's configure
								ServerConfig remoteServer = ServerConfig.getServerConfig(currentConfig.rootContext,
										remoteBranch, true);

								Socket socket = new Socket(remoteServer.localhostIP, Integer
										.parseInt(remoteServer.UDPPort));
								socket.setSoTimeout(BankUtility.WaitingLength);

								PacketWrapper.writePacket(socket, transferPacket);
								receivePacket = PacketWrapper.readPacket(socket);

							}
							catch (Exception er) {
								done = false;
								receivePacket.status = BankUtility.TRANSFER_REJECTED;
							}

							if (receivePacket.status == BankUtility.TRANSFER_PREPARED) {
								done = true;
								// Transfer succeeded, check the data file
								int localAcnt = logPacket.source % 10000;
								dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
								float oldValue = dataFile.readFloat();
								// Check if the data match the log
								if (oldValue == logPacket.oldValue) {
									// Write new value to the account
									dataFile.seek(localAcnt * BankUtility.ACCOUNT_SIZE);
									dataFile.writeFloat(logPacket.newValue);
								}
							}
							else {
								// Transfer failed, do nothing or retry
							}
						}// End of while

					}
				}
			}
			catch (Exception e) {
				// Error happened in processing the log packet, exit
				System.out.println("Error happened in processing the log packet, exit.");
				Runtime.getRuntime().exit(1);
			}
		}
		System.out.println("Checking log file and data file succeeded.");
	}

}
